
/*********************************************************************************
 *  Copyright© 2010 Gabriele Valentini (gabriele.valentini@mail.polimi.it)	 *
 *										 *
 *  This file is part of Evoptool.						 *
 * 										 *
 *  Evptool is free software: you can redistribute it and/or modify              *
 *  it under the terms of the GNU General Public License as published by         *
 *  the Free Software Foundation, either version 3 of the License, or            *
 *  (at your option) any later version.                                          *
 *										 *
 *  Evoptool is distributed in the hope that it will be useful,                  *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of  		 *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 		 *
 *  GNU General Public License for more details.				 *
 *										 *
 *  You should have received a copy of the GNU General Public License		 *
 *  along with Evoptool.  If not, see <http://www.gnu.org/licenses/>.    	 *	
 *********************************************************************************/
 
#include <algorithm>
#include <float.h>
#include <list>

#include "COMIT.h"

/* Class constructor. */
COMIT::COMIT(unsigned int popSize, unsigned int samples, unsigned int max_influence, const gsl_rng* rnd) : 
	EDA(popSize, rnd), 
	_samples(samples),
	_max_influence(max_influence) {
	
}

/* Class destructor. */
COMIT::~COMIT() {
	for(unsigned int i=0; i<_pop.size(); i++)
		delete _pop[i];
	_pop.clear();
}

/* Set the reference to the given objective function and initialize a random population. */
void COMIT::setObjectiveFunction(ObjectiveFunction* of) {
	Algorithm::setObjectiveFunction(of);

	for (unsigned int i=0;i<_popSize;i++) {
		BinaryInstance* instance = new BinaryInstance(objectiveFunctionSize_);
		for (unsigned int j=0;j<objectiveFunctionSize_;j++) {
			instance->set(j,gsl_ran_bernoulli(_rnd, 0.5));
		}
		_pop.push_back(instance);
	}
}

/* This function defines the single generation step of the algorithm. */
void COMIT::run() {
	estimation();
	sampling();
	selection();
	fastsearch();
	replace();
}

/* Build an optimal dependency tree */
void COMIT::estimation() {
	int best_match_in_tree[objectiveFunctionSize_];
	list<int> bit_idx;
	
	/*Initialize bit_idx list with a random permutation of indexes */
	vector<int> bit_idx_support;
	for (unsigned int j=0; j<objectiveFunctionSize_; j++)
		bit_idx_support.push_back(j);
	
	random_shuffle(bit_idx_support.begin(), bit_idx_support.end()); 
	
	for (unsigned int j=0; j<objectiveFunctionSize_; j++)
		bit_idx.push_back(bit_idx_support[j]);
	
	bit_idx_support.clear();
	
	/* Creating root */
	
	_bnn_root = new BNN(bit_idx.front());
	_bayesian_map[_bnn_root->getIndex()] = _bnn_root;
	
	for (unsigned int j=0; j<objectiveFunctionSize_; j++) 
		best_match_in_tree[j] = bit_idx.front();
	
	bit_idx.remove(best_match_in_tree[0]);
	
	/* Building rest of tree */ 
	while (bit_idx.size() != 0) {
		int selected = bit_idx.front();
		double max =  mutualInformation(selected, best_match_in_tree[selected]);
		
		/* Select bit index with maximum mutual information */
		for (list<int>::iterator it=bit_idx.begin(); it != bit_idx.end(); it++) {
			double temp = mutualInformation(*it, best_match_in_tree[*it]);
			if (temp > max) {
				max = temp;
				selected = *it;
			}
		}
		bit_idx.remove(selected);

		/* Add selected index to the tree */
		BNN *parent = _bayesian_map[best_match_in_tree[selected]];
		if (parent == NULL) 	cerr << "BNN::find() ERROR!" << endl; 
		
		BNN *node  = new BNN(selected);
		parent->addSon(selected);
		_bayesian_map[selected] = node;
			
		/* Update best_match_in_tree informations */
		if(bit_idx.size() != 0) {
			for (list<int>::iterator it=bit_idx.begin(); it != bit_idx.end(); it++) { 
				if(mutualInformation(*it, best_match_in_tree[*it]) < mutualInformation(*it, selected))
					best_match_in_tree[*it] = selected;
			} 
		}
	}
}

/* Return the mutual information between idx1 and idx2 bits */
double COMIT::mutualInformation(int idx1, int idx2) {
	double p1 = 0, p2 = 0, value = 0;
	double p12[4] = {0,0,0,0};
	
	/* Calculating empirical probabilities */
	for (unsigned int i=0; i<_popSize; i++) {
		if (_pop[i]->get(idx1) == 1) p1++;
		if (_pop[i]->get(idx2) == 1) p2++;
		if ((_pop[i]->get(idx1) == 0) && (_pop[i]->get(idx2) == 0)) p12[0]++;
		if ((_pop[i]->get(idx1) == 0) && (_pop[i]->get(idx2) == 1)) p12[1]++;
		if ((_pop[i]->get(idx1) == 1) && (_pop[i]->get(idx2) == 0)) p12[2]++;
		if ((_pop[i]->get(idx1) == 1) && (_pop[i]->get(idx2) == 1)) p12[3]++;
	}
	
	p1 /= _popSize;
	p2 /= _popSize;
	for (int i=0; i<4; i++)
		p12[i] /= _popSize;
	
	/* Calculating mutual information */
	if (p12[0] != 0) {
		if ((1 - p1 == 0) || (1 - p2 == 0))	value += p12[0] * log(DBL_MAX);
        	else	value += p12[0] * log(p12[0] / ((1 - p1) * (1 - p2)));
        }
        
        if (p12[1] != 0) {
		if ((1 - p1 == 0) || (p2 == 0))	value += p12[1] * log(DBL_MAX);
        	else	value += p12[1] * log(p12[1] / ((1 - p1) * p2));
        }
        
        if (p12[2] != 0) {
		if ((p1 == 0) || (1 - p2 == 0))	value += p12[2] * log(DBL_MAX);
        	else	value += p12[2] * log(p12[2] / (p1 * (1 - p2)));
        }
        
        if (p12[3] != 0) {
		if ((p1 == 0) || (p2 == 0))	value += p12[3] * log(DBL_MAX);
        	else	value += p12[3] * log(p12[3] / (p1 * p2));
        }
	
	return value;
}

/* Generate _samples from the dependency tree */
void COMIT::sampling() {
	double empirical_probability = 0;
	
	/* Calculating empirical probability for root */
	for (unsigned int j=0; j<_popSize; j++)
		if (_pop[j]->get(_bnn_root->getIndex()) == 1) 	empirical_probability++;
	empirical_probability /= _popSize;
	
	/* Generating new samples */
	for (unsigned int i=0; i<_samples; i++) {
		int value;
		BinaryInstance *instance = new BinaryInstance(objectiveFunctionSize_);
		
		/* Setting first bit of new individual */
		value = gsl_ran_bernoulli(_rnd, empirical_probability);
		_bnn_root->setValue(value);
		instance->set(_bnn_root->getIndex(), value);
		
		/* Setting others bits */
		recursiveSampling(instance, _bnn_root->getIndex());
		
		_sample_vec.push_back(instance);
	}

	/* Delete previous tree */
	if (_bayesian_map.size() != 0) {
		for(unsigned int i=0; i<_bayesian_map.size(); i++)
			delete _bayesian_map[i];
		_bayesian_map.clear();
	}
}

/* Generate remaining part of a sample visiting the bayesian network tree */
void COMIT::recursiveSampling(BinaryInstance *instance, int root_idx) {
	vector<int> son_idx;
	vector<double> son_cprob; 
	BNN *root;
	
	root = _bayesian_map[root_idx];
	son_idx = root->getSon();
	
	if (son_idx.size() != 0) {
		/* Calculating conditional probabilities of sons */
		
		son_cprob = conditionalProbabilities(son_idx, root->getIndex(), root->getValue());
		
		/* Setting values */
		for(unsigned int i=0; i<son_idx.size(); i++) {
			int value = gsl_ran_bernoulli(_rnd, son_cprob[i]);
			_bayesian_map[son_idx[i]]->setValue(value);
			instance->set(son_idx[i], value);
		}
		
		/* Sampling recursively all the root sons */
		for(unsigned int i=0; i<son_idx.size(); i++)
			recursiveSampling(instance, son_idx[i]);
	}
}


/* Calculate the conditional probabilities for idx sons of root_idx, knowing value of root_idx */
vector<double> COMIT::conditionalProbabilities(vector<int> idx, int root_idx, int root_value) {
	vector<double> cond_prob;
	int num_root = 0;
	
	for (unsigned int k=0; k<idx.size(); k++)
		cond_prob.push_back(0);
	
	for (unsigned int j=0; j<_popSize; j++) {
		if(_pop[j]->get(root_idx) == root_value) {
			num_root++;
			for (unsigned int i=0; i<idx.size(); i++)
				if(_pop[j]->get(idx[i]) == 1) cond_prob[i]++;
		}
	}
	
	for (unsigned int i=0; i<cond_prob.size(); i++)
		cond_prob[i] /= num_root;
	
	return cond_prob;
}


/* Replace up to max_influence individuals from original population with new coming from fastsearch */
void COMIT::replace() {
	
	_pop.insert(_pop.end(), _new_fast_search.begin(), _new_fast_search.end());
	std::sort(_pop.begin(), _pop.end(), Comparator(this));
	int ns = _pop.size();
	
	for (int j=_popSize; j<ns; j++)
		delete _pop[j];
		
	_pop.erase(_pop.begin() + _popSize, _pop.end());
	_new_fast_search.clear();
}
