
/*********************************************************************************
 *  Copyright© 2010 Gabriele Valentini (gabriele.valentini@mail.polimi.it)	 *
 *										 *
 *  This file is part of Evoptool.						 *
 * 										 *
 *  Evptool is free software: you can redistribute it and/or modify              *
 *  it under the terms of the GNU General Public License as published by         *
 *  the Free Software Foundation, either version 3 of the License, or            *
 *  (at your option) any later version.                                          *
 *										 *
 *  Evoptool is distributed in the hope that it will be useful,                  *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of  		 *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 		 *
 *  GNU General Public License for more details.				 *
 *										 *
 *  You should have received a copy of the GNU General Public License		 *
 *  along with Evoptool.  If not, see <http://www.gnu.org/licenses/>.    	 *	
 *********************************************************************************/

#include "DEUM.h"
#include "SVD.h"

/* Class costructor. */
DEUM::DEUM(unsigned int pop_size, double cooling_rate, int sampler_iterations, const gsl_rng* rnd) :
	EDA(pop_size,rnd),
	_cooling_rate(cooling_rate),
	_sampler_iterations(sampler_iterations) {

}

/* Class costructor with a given population and MRF. */
DEUM::DEUM(vector<BinaryInstance*> pop, vector<Clique*> mrf, ObjectiveFunction *of, double cooling_rate, int sampler_iterations, const gsl_rng* rnd) :
	EDA(pop.size(), rnd),
	_cooling_rate(cooling_rate),
	_sampler_iterations(sampler_iterations) {
	Algorithm::setObjectiveFunction(of);
	_pop = pop;
	_mrf = mrf;
	_alpha = gsl_vector_alloc(_mrf.size() + 1);
}

/* Class destructure. */
DEUM::~DEUM() {
	unsigned int i; 
	
	if (_alpha != NULL) gsl_vector_free(_alpha);
	for (i = 0; i < _pop.size(); i++)
		delete _pop[i];
	_pop.clear();
	for (i = 0; i < _mrf.size(); i++)
		delete _mrf[i];
	_mrf.clear();
}

/* Sets the objective function for this execution and initialize algorithm structure. */
void DEUM::setObjectiveFunction(ObjectiveFunction* of) {
	Algorithm::setObjectiveFunction(of);
	
	setupStructure();
	checkPopSize();
	
	for (unsigned int i=0;i<_popSize;i++) {
		BinaryInstance* instance = new BinaryInstance(objectiveFunctionSize_);
		for (unsigned int j=0;j<objectiveFunctionSize_;j++) {
			instance->set(j,gsl_ran_bernoulli(_rnd, 0.5));
		}
		_pop.push_back(instance);
	}
	computeParameters();
}

/* Performs a single generation of the algorithm and returns the sampled individual. */
BinaryInstance* DEUM::gedaGeneration() {
	int selected; 
	BinaryInstance *sampled; 
	
	estimation();
	selected = gsl_rng_uniform_int(_rnd, _pop.size());
	sampled = gibbsKPCoolingSampler(_pop[selected], _sampler_iterations, _cooling_rate);
	return sampled; 
}

/* Samples the new population using gibbs sampler. It produce a population double of the original one
 * so in the next step is possible to select at 50% maintainig the minimum size for use SVD. */
void DEUM::sampling() {
	unsigned int i; 
	BinaryInstance *son1, *son2; 
	vector<BinaryInstance*> tmp; 
	
	/* Uncomment to print the MRF parameters vector. */
	//printModel();
	
	/* Sampling new population (double size). */
	for (i = 0; i < _popSize; i++) {
		son1 = gibbsKPCoolingSampler(_pop[i], _sampler_iterations, _cooling_rate);
		son2 = gibbsKPCoolingSampler(_pop[i], _sampler_iterations, _cooling_rate);
		tmp.push_back(son1);
		tmp.push_back(son2);
	}
	
	/* Deleting old population. */
	for (i = 0; i < _pop.size(); i++)
		delete _pop[i];
	
	/* Saving new population. */
	_pop.clear();
	_pop.assign(tmp.begin(), tmp.end());
	tmp.clear();
}

/* Performs truncation selection at 50% of population. This is done by maintaing the original 
 * population size and by sampling the double of that size. */
void DEUM::selection() {
	unsigned int i; 
	
	std::sort(_pop.begin(), _pop.end(), Comparator(this));
	for( i = _popSize; i < _pop.size(); i++)
		delete _pop[i];
	_pop.erase(_pop.begin() + _popSize, _pop.end());
}

/* Estimates the MRF distribution for this population. */
void DEUM::estimation() {
	computeParameters();
}

/* Predicts the fitness of an individual by inverting the distribution equation -ln(f(x)) = U(x). During 
 * estimation we compute MRF parameters by knowing the fitness, now we estimate the fitness by knowing the 
 * MFR parameters. */
double DEUM::predictFitness(BinaryInstance *instance) {
	double value = 0, f; 
		
	/* Compute the sum of the potential energy of the cliques. */
	for (unsigned int j = 0; j < _mrf.size(); j++) {
		value += gsl_vector_get(_alpha, j) * _mrf[j]->harmonicalValue(instance);	
	}
	
	value += gsl_vector_get(_alpha, _mrf.size());	// Add constant
	value *= -1;
	value /= TEMPERATURE; 
	f = exp(value) - 2; 					// Constant added (-ln(f(x) - 2)) 
	
	return f; 
}

/* Computes the singular value decomposition for solving the linear system in order to estimate
 * the MRF coefficients. The result is stored in the global variable _alpha. These function works 
 * with every kind of MRF structure. (SVD using gsl)*/
//void DEUM::computeParameters() {
//	double fitness, tmp; 
//	int i,j,row, col; 
//	gsl_matrix *AU, *V;
//	gsl_vector *S, *b, *work; 
//	
//	/* An alpha for each clique and one for the constant. */
//	col = _mrf.size() + 1;
//	row = _popSize; 
//	
//	/* Initializing matrixes and vectors. */
//	AU = gsl_matrix_calloc(max(row,col), col);
//	b = gsl_vector_calloc(max(row,col));
//	V = gsl_matrix_alloc(col, col);
//	S = gsl_vector_alloc(col);
//	work = gsl_vector_alloc(col);
//	
//	/* Initializing structures. */
//	for (i = 0; i < row; i++) {
//	
//		fitness = _pop[i]->validCache() ? _pop[i]->getFitnessCache() : _of->f(_pop[i]);
//		tmp = -1 * TEMPERATURE * log(fitness + 2);
//		gsl_vector_set(b,i,tmp);
//		
//		for (j = 0; j < col - 1; j++) {
//			tmp = _mrf[j]->harmonicalValue(_pop[i]);
//			gsl_matrix_set(AU,i,j,tmp);
//		}
//		
//		gsl_matrix_set(AU,i,col - 1,1);
//	}
//	
//	/* Solving linear system. */
//	gsl_linalg_SV_decomp(AU, V, S, work);
//	
//	/* Remove negligable value. (S = w) */
//	double smin = gsl_vector_max(S) * 1E-6;
//	for (i = 0; i < col; i++) {
//		if (gsl_vector_get(S,i) < smin)	{
//			gsl_vector_set(S,i,0.0);
//		}
//	}
//	
//	gsl_linalg_SV_solve(AU, V, S, b, _alpha);
//	
//	/* Freeing memory. */
//	gsl_vector_free(S);
//	gsl_vector_free(b);
//	gsl_vector_free(work);
//	gsl_matrix_free(AU);
//	gsl_matrix_free(V); 
//}

/* SVD from Numerical Recipies */
void DEUM::computeParameters() {
	double f, tmp; 
	int i,j,row, col, mrc; 
	double **AU, **V;
	double *S, *b, *x; 
	
	/* An alpha for each clique and one for the constant. */
	col = _mrf.size() + 1;
	row = _popSize; 
	mrc = max(row,col);

	/* Initializing matrixes and vectors. */
	AU = new double*[mrc];
	b = new double[mrc];
	V = new double*[col];
	S = new double [col];
	x = new double [col];
	
	for (j = 0; j < col; j++) V[j] = new double[col];
	
	/* Initializing structures. */
	for (i = 0; i < row; i++) {
		AU[i] = new double[col];
		
		f = _pop[i]->validCache() ? _pop[i]->getFitnessCache() : fitness(_pop[i]);
		tmp = -1 * TEMPERATURE * log(f + 2);
		b[i] = tmp;
		 
		for (j = 0; j < col - 1; j++) {
			tmp = _mrf[j]->harmonicalValue(_pop[i]);
			AU[i][j] = tmp;
		}

		AU[i][col - 1] = 1;
	}
	
	/* Padding AU and b if necessary. */
	if (row < col) {
		for (i = row; i < col; i++) {
			AU[i] = new double[col];
			for (j = 0; j < col; j++) AU[i][j] = 0.0;
			b[i] = 0.0;
		}
	}
	
	/* Solving linear system. */
	SVD::svd(AU, S, V, mrc, col);
	
	/* Remove negligable value. (S = w) */
	double smin = 1E-20;
	for (i = 0; i < col; i++) {
		if (S[i] > smin) smin = S[i];	
	}
	
	smin *= 1E-6;
	SVD::zeroNegligables(S, col, smin);
	
	x = SVD::svbksb(AU, S, V, mrc, col, b);
	
	for (i = 0; i < col; i++) {
		gsl_vector_set(_alpha, i, x[i]);
	}
	
	delete[] x;
	delete[] b;
	delete[] S;
	delete[] AU;
	delete[] V;
}

/* Samples a new individual starting from a previous one, for a fixed number of iterations and returns it. */
BinaryInstance* DEUM::gibbsKPCoolingSampler(BinaryInstance *instance, int maxiterations, double cooling_rate) {
	int i, bit; 
	double temperature = TEMPERATURE, p;
	BinaryInstance *sampled = new BinaryInstance(*instance);
	
	/* Iteratively sample the same individual. */
	for (i = 0; i < maxiterations; i++) {
		temperature *= cooling_rate; 
		bit = gsl_rng_uniform_int(_rnd, objectiveFunctionSize_);
		p = marginalProbability(sampled, bit, temperature);
		sampled->set(bit, gsl_ran_bernoulli(_rnd, p));
	}
	
	return sampled; 
}

/* Returns the marginal probability for this allele as sum of potential energies of cliques that 
 * contains the allele. */
double DEUM::marginalProbability(BinaryInstance *instance, int bit, double temperature) {
	double value = 0, p; 
		
	/* Compute the sum of the potential energy of the cliques. */
	for (unsigned int j = 0; j < _mrf.size(); j++) {
		if (_mrf[j]->find(bit)) {
			value += gsl_vector_get(_alpha, j) * _mrf[j]->harmonicalValue(instance);	
		}
	}		
	
	value *= instance->getHarmonicEncoding(bit);	
	p = exp(2 * value / temperature);
	p = 1 / (1 + p);
	return p; 
}

/* Generates univariate cliques for every position of the individual and adds their to the 
 * MRF representation. */
void DEUM::genUnivariates() {
	unsigned int i; 
	Clique *univariate; 
	
	for (i = 0; i < objectiveFunctionSize_; i++) {
		univariate = new Clique();
		univariate->add(i);
		_mrf.push_back(univariate);
	}
}

/* Generates bivariate cliques for every position of the individual as a bivariate chain 
 * with rings (i, i + offset). */
void DEUM::genBivariates(unsigned int offset) {
	unsigned int i; 
	Clique *bivariate; 
	
	offset = offset % objectiveFunctionSize_;
	for (i = 0; i < objectiveFunctionSize_; i++) {
		bivariate = new Clique();
		bivariate->add(i);
		if (i < objectiveFunctionSize_ - offset) 
			bivariate->add(i + offset);
		else 
			bivariate->add(i + offset - objectiveFunctionSize_);
		_mrf.push_back(bivariate);
	}
}

/* Generates multivariate clique for every position of the individual as specified from 
 * parameters. Offset is the array of offsets and size the size of the array. The generated
 * clique are in the form (size = 3; offset{1,2,3}; clique(i, i + 1, i + 2; i + 3)). */
void DEUM::genMultivariates(unsigned int size, unsigned int* offset) {
	unsigned int i, j; 
	Clique *multivariate;
	
	for (i = 0; i < objectiveFunctionSize_; i++) {
		multivariate = new Clique();
		multivariate->add(i);
		for (j = 0; j < size; j++) {
			if (i < objectiveFunctionSize_ - offset[j])
				multivariate->add(i + offset[j]);
			else
				multivariate->add(i + offset[j] - objectiveFunctionSize_);
		}
		_mrf.push_back(multivariate);
	}
}

/* Checks and eventually fixes the population size for assure the minimum number of individuals
 * for solve the linear system with SVD. */
void DEUM::checkPopSize() {
	_popSize = _popSize >= _mrf.size() + 1 ? _popSize : _mrf.size() + 1; 
}

void DEUM::printModel() {
	for (unsigned int i = 0; i < _mrf.size(); i++)
		cout << "(" << _mrf[i]->toString() << "," << gsl_vector_get(_alpha,i) << ")";
	cout << "(C = "<< gsl_vector_get(_alpha, _mrf.size()) << ")" <<endl << endl;
}

/* Returns the algorithm name. */
string DEUM::getName() {
	return "DEUM";
}
