
/*********************************************************************************
 *  Copyright© 2010 Gabriele Valentini (gabriele.valentini@mail.polimi.it)	 *
 *										 *
 *  This file is part of Evoptool.						 *
 * 										 *
 *  Evptool is free software: you can redistribute it and/or modify              *
 *  it under the terms of the GNU General Public License as published by         *
 *  the Free Software Foundation, either version 3 of the License, or            *
 *  (at your option) any later version.                                          *
 *										 *
 *  Evoptool is distributed in the hope that it will be useful,                  *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of  		 *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 		 *
 *  GNU General Public License for more details.				 *
 *										 *
 *  You should have received a copy of the GNU General Public License		 *
 *  along with Evoptool.  If not, see <http://www.gnu.org/licenses/>.    	 *	
 *********************************************************************************/

#ifndef DEUM_H_
#define DEUM_H_

#include "EDA.h"
#include "Clique.h"

#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_errno.h>
#include <algorithm>

#define TEMPERATURE 1
#define SAMPLER_ITERATIONS _popSize

using namespace std;

/* This class defines an abstract object for the DEUM algorithms family, where there are implemented all the 
 * common function for a DEUM algorithm with a fixed structure. */

/* In order to solve the linear sistem of equations with SVD is required that the population size will be great 
 * almost as the numeber of cliques plus one (for the costant term). For assure that, a population size checking 
 * is performed as last operation during algorithm initialization. */
 
class DEUM : public EDA {

	protected:
		vector<Clique*> _mrf;		/* The structure of the MRF as vector of cliques. 			*/
		gsl_vector *_alpha;		/* List of clique coefficients computed plus costant term. 		*/
		double _cooling_rate;  		/* Cooling rate for the gibbs sampler (near 0 is fast, near 1 is slow). */
		int _sampler_iterations; 	/* Sampler iterations. 							*/
		
	public:
		DEUM(unsigned int pop_size, double cooling_rate, int sampler_iterations, const gsl_rng* rnd);
		DEUM(vector<BinaryInstance*> pop, vector<Clique*> mrf, ObjectiveFunction *of, double cooling_rate, int sampler_iterations, const gsl_rng* rnd);
		virtual ~DEUM();

		void sampling();
		void selection();
		void estimation();

		string getName();
		void printModel();
		void setObjectiveFunction(ObjectiveFunction* of);
		BinaryInstance* gedaGeneration();
	
	protected: 
		double marginalProbability(BinaryInstance *instance, int bit, double temperature);
		double predictFitness(BinaryInstance *instance);
		void computeParameters();
		BinaryInstance* gibbsKPCoolingSampler(BinaryInstance *instance, int maxiterations, double cooling_rate);
		
		/* Setup the MRF structure and initialize _alpha vector. */
		virtual void setupStructure() = 0; 
		void genUnivariates();
		void genBivariates(unsigned int offset = 1);
		void genMultivariates(unsigned int size, unsigned int* offset);
		void checkPopSize();
};
#endif
