
/*********************************************************************************
 *  Copyright© 2010 Gabriele Valentini (gabriele.valentini@mail.polimi.it)	 *
 *										 *
 *  This file is part of Evoptool.						 *
 * 										 *
 *  Evptool is free software: you can redistribute it and/or modify              *
 *  it under the terms of the GNU General Public License as published by         *
 *  the Free Software Foundation, either version 3 of the License, or            *
 *  (at your option) any later version.                                          *
 *										 *
 *  Evoptool is distributed in the hope that it will be useful,                  *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of  		 *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 		 *
 *  GNU General Public License for more details.				 *
 *										 *
 *  You should have received a copy of the GNU General Public License		 *
 *  along with Evoptool.  If not, see <http://www.gnu.org/licenses/>.    	 *	
 *********************************************************************************/
 
#include "pbilCOMIT.h"
#include <algorithm>

/* Class constructor. */
pbilCOMIT::pbilCOMIT(unsigned int popSize, unsigned int samples, unsigned int max_influence, double gamma, double mut_rate, int percentage_selection, int iterations, const gsl_rng* rnd) :
	COMIT(popSize, samples, max_influence, rnd),
	_gamma(gamma),
	_mut_rate(mut_rate),
	_percentage_selection(percentage_selection),
	_iterations(iterations) {

}

/* Class destructor. */
pbilCOMIT::~pbilCOMIT() {

}

/* Selection operator*/
void pbilCOMIT::selection() {
	int new_num = (int) _sample_vec.size() * _percentage_selection / 100;
	std::sort(_sample_vec.begin(), _sample_vec.end(), Comparator(this));
	_new_fast_search.clear();
	_new_fast_search.insert(_new_fast_search.end(), _sample_vec.begin(), _sample_vec.begin() + new_num);
	
	/* Deleting samples from _sample-vec vector */
	for (unsigned int i=new_num; i<_sample_vec.size(); i++)
		delete _sample_vec[i];
	if (_sample_vec.size() != 0)	_sample_vec.clear(); 
} 

/* PBIL based fast-search operator. */
void pbilCOMIT::fastsearch() {
	/* Initialize probability vector if necessary */
	if(_prob.size() == 0)
		for (unsigned int i=0; i<objectiveFunctionSize_; i++)
			_prob.push_back(0);
		
	/* Run PBIL algorithm for a fixed number of iterations */
	for(int i=0; i<_iterations; i++) {
		double sum;
		
		/* Estimate probability vector */
		for (unsigned int j=0; j<_prob.size(); j++) {
			sum = 0;
			for (unsigned int i=0; i<_new_fast_search.size(); i++) {
				BinaryInstance *instance;
				instance = _new_fast_search[i];
				sum += instance->get(j);
			}
			_prob[j] = (1-_gamma)*_prob[j] + _gamma*(sum/_new_fast_search.size());
		}
		
		/* Mutate probability vector */
		for (unsigned int i=0; i<_prob.size(); i++) {
			if(gsl_ran_bernoulli(_rnd, _mut_rate) == 1) {
				if(_prob[i] > 0.5)
					_prob[i] *= 0.95;
				else
					_prob[i] = _prob[i]*0.95 + 0.05;  
			}
		}
		
		/* Sampling probability vector (generate 2*max_influence samples) */
		for (unsigned int i=0; i<2*_max_influence; i++) {
			BinaryInstance *instance = new BinaryInstance(_prob.size());
			for (unsigned int j=0; j<_prob.size(); j++) {
				instance->set(j,gsl_ran_bernoulli(_rnd, _prob[j]));
			} 
			_new_fast_search.push_back(instance);
		}
		
		/* Truncation selection at 50% */
		std::sort(_new_fast_search.begin(), _new_fast_search.end(), Comparator(this));
		for (unsigned int i=_max_influence ; i< _new_fast_search.size(); i++)
			delete _new_fast_search[i];
		_new_fast_search.erase(_new_fast_search.begin() + _max_influence, _new_fast_search.end());
	}
}

/* Returns the algorithm name. */
string pbilCOMIT::getName() {
	return "pbilCOMIT";
}
