
/*********************************************************************************
 *  Copyright© 2010 Gabriele Valentini (gabriele.valentini@mail.polimi.it)	 *
 *										 *
 *  This file is part of Evoptool.						 *
 * 										 *
 *  Evptool is free software: you can redistribute it and/or modify              *
 *  it under the terms of the GNU General Public License as published by         *
 *  the Free Software Foundation, either version 3 of the License, or            *
 *  (at your option) any later version.                                          *
 *										 *
 *  Evoptool is distributed in the hope that it will be useful,                  *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of  		 *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 		 *
 *  GNU General Public License for more details.				 *
 *										 *
 *  You should have received a copy of the GNU General Public License		 *
 *  along with Evoptool.  If not, see <http://www.gnu.org/licenses/>.    	 *	
 *********************************************************************************/

#include "Evoptool.h"
#include "GraphicsTool.h"
#include "XMLParser.h"

#include <random_seed.h>
#include <string.h>
#include <iostream>
#include <fstream>
#include <sys/stat.h>


/*-----------------------------------------------------------------------------------------------------------------*/
/*						PUBLIC FUNCTIONS 						   */
/*-----------------------------------------------------------------------------------------------------------------*/

/* Class constructor. */
Evoptool::Evoptool(ExeMode mode) : 
	_mode(mode) {

}

/* Class destructor. */
Evoptool::~Evoptool() {

}

/* Initializes the tool: thread system, log file, signal handlers, temporany files. */
void Evoptool::init() {
	bool rval; 
	
	/* Initialize support variables. */
	_bestAsImageFilesAllocated = false;
	
	/* Initialize thread system. */
	if(!Glib::thread_supported()) Glib::thread_init();

	/* Delete old log file by truncating it. */
	ofstream logFile(LOG_FILE, ios_base::trunc);
	if (logFile.fail()) cerr << "LogManager: Error while opening log file." << endl;
	logFile.close();
	
	/*Connecting signal handlers. */
	signalError().connect(sigc::mem_fun(*this, &Evoptool::signalErrorHandler));
	if (_mode == GUI) {
		signalInitializeTest().connect(sigc::mem_fun(*this, &Evoptool::signalInitializeTestHandler));
		signalRunTest().connect(sigc::mem_fun(*this, &Evoptool::signalRunTestHandler));
		signalFinalizeTest().connect(sigc::mem_fun(*this, &Evoptool::signalFinalizeTestHandler));
		signalStopTest().connect(sigc::mem_fun(*this, &Evoptool::signalStopTestHandler));
		signalComputeStatistics().connect(sigc::mem_fun(*this, &Evoptool::signalComputeStatisticsHandler));	
	}
	
	/* Deleting temporany files. */
	rval = cleanTemporanyFiles();
	if (!rval) signalError().emit(Evoptool::UNRECOVERABLE_ERROR);
}

/* This functions perform a complete test specified by a parameters file, including the computation of
 * final statistics and the exportation in a tar archive if requested. */
void Evoptool::execute(const char *parametersFile) {
	bool rval;
	Glib::Thread *runnerThread;
	
	printLog((char*) __FILE__, __LINE__, (char*) "Starting execution in command line mode.", Evoptool::MESSAGE);
	
	printLog((char*) __FILE__, __LINE__, (char*) "Parsing parameters file.", Evoptool::MESSAGE);

	_testParameters = new TestParameters();
	initTestParameters(_testParameters);
	rval = XMLParser::parse(_testParameters, parametersFile);
	
	if (!rval) {
		printLog((char*) __FILE__, __LINE__, (char*) "Cannot load parameters file.", Evoptool::ERROR);
		signalError().emit(Evoptool::UNRECOVERABLE_ERROR);
	}
	
	printLog((char*) __FILE__, __LINE__, (char*) "Configuring statistics computation.", Evoptool::MESSAGE);
	rval = configureStatistics();
	
	if (!rval) signalError().emit(Evoptool::UNRECOVERABLE_ERROR);
	printLog((char*) __FILE__, __LINE__, (char*) "Statistics computation configured.", Evoptool::MESSAGE);
	
	printLog((char*) __FILE__, __LINE__, (char*) "Parameters loaded.", Evoptool::MESSAGE);
	printLog((char*) __FILE__, __LINE__, (char*) "Running the test.", Evoptool::MESSAGE);
	
	runnerThread = Glib::Thread::create(sigc::mem_fun(*this, &Evoptool::runnerThread),true);
	runnerThread->join();
	
	printLog((char*) __FILE__, __LINE__, (char*) "Test done.", Evoptool::MESSAGE);
	printLog((char*) __FILE__, __LINE__, (char*) "Finalizing test.", Evoptool::MESSAGE);
	
	printLog((char*) __FILE__, __LINE__, (char*) "Computing final statistics.", Evoptool::MESSAGE);
	if (!computeStatistics()) {
		printLog((char*) __FILE__, __LINE__, (char*) "Error during statistics computing.", Evoptool::ERROR);
		signalError().emit(Evoptool::UNRECOVERABLE_ERROR);
	}
	printLog((char*) __FILE__, __LINE__, (char*) "Final statistics computed.", Evoptool::MESSAGE);
		
	if (_testParameters->exportData) {
		rval = exportDatas(_testParameters->destinationFolder, _testParameters->archiveName);
		if (!rval) signalError().emit(Evoptool::UNRECOVERABLE_ERROR);
	}
	
	printLog((char*) __FILE__, __LINE__, (char*) "Test finalized.", Evoptool::MESSAGE);
	printLog((char*) __FILE__, __LINE__, (char*) "Cleaning up tool's structures.", Evoptool::MESSAGE);
	
	deleteTestParameters();
	printLog((char*) __FILE__, __LINE__, (char*) "Execution done, exit.", Evoptool::MESSAGE);
}

/* Public wrapper for allocRNG function. Allows the GUI to initialize TestParameters structure. */
gsl_rng* Evoptool::Evoptool::getRNG() {
	return allocRNG();
}
/* ------------------------------------------------- Signal accessors -------------------------------------------- */

/* Signal Error accessor. */
Evoptool::SignalError Evoptool::signalError() {
	return _signalError;
}

/* Signal Initialize Test accessor. */
Evoptool::SignalInitializeTest Evoptool::signalInitializeTest() {
	return _signalInitializeTest;
}

/* Signal Run Test accessor. */
Evoptool::SignalRunTest Evoptool::signalRunTest() {
	return _signalRunTest;
}

/* Signal Finalize Test accessor. */
Evoptool::SignalFinalizeTest Evoptool::signalFinalizeTest() {
	return _signalFinalizeTest;
}

/* Signal Test Executed accessor. */
Evoptool::SignalTestExecuted Evoptool::signalTestExecuted() {
	return _signalTestExecuted;
}

/* Signal Stop Test accessor. */
Evoptool::SignalStopTest Evoptool::signalStopTest() {
	return _signalStopTest;
}

/* Signal Compute Statistics accessor. */
Evoptool::SignalComputeStatistics Evoptool::signalComputeStatistics() {
	return _signalComputeStatistics;
}

/* Signal new log accessor. */
Evoptool::SignalNewLog Evoptool::signalNewLog() {
	return _signalNewLog;
}
	
/*-----------------------------------------------------------------------------------------------------------------*/
/* 						PRIVATE FUNCTIONS			       			   */
/*-----------------------------------------------------------------------------------------------------------------*/

/* ---------------------------------------- Signal in-class handlers --------------------------------------------- */

/* This handler set the TestParameters structure for the test and runs the initializer thread. */
void Evoptool::signalInitializeTestHandler(TestParameters *parameters) {
	_testParameters = parameters;
	Glib::Thread::create(sigc::mem_fun(*this, &Evoptool::initializerThread), false);
	printLog((char*) __FILE__, __LINE__, (char*) "Signal initialize test processed.", Evoptool::MESSAGE);
}

/* This handler runs the runner thread. */
void Evoptool::signalRunTestHandler() {
	Glib::Thread::create(sigc::mem_fun(*this, &Evoptool::runnerThread), false);
	printLog((char*) __FILE__, __LINE__, (char*) "Signal run test processed.", Evoptool::MESSAGE);
}

/* This handler runs the finalizer thread. */
void Evoptool::signalFinalizeTestHandler() {
	Glib::Thread::create(sigc::mem_fun(*this, &Evoptool::finalizerThread), false);
	printLog((char*) __FILE__, __LINE__, (char*) "Signal finalize test processed.", Evoptool::MESSAGE);
}

/* This handler runs the stopper thread. */
void Evoptool::signalStopTestHandler() {
	Glib::Thread::create(sigc::mem_fun(*this, &Evoptool::stopperThread), false);
	printLog((char*) __FILE__, __LINE__, (char*) "Signal stop test processed.", Evoptool::MESSAGE);
}

/* This handler wrap the statistics computation call for the GUI. */
void Evoptool::signalComputeStatisticsHandler(unsigned int options, unsigned int height, unsigned int width ) {
	computeStatistics(options, height, width);
}

/* This handler tries to repair to the signaled error and recover the execution of the tool. If is not possible
 * recover the execution, the handler will abort the tool execution. */
void Evoptool::signalErrorHandler(EvoptoolError error) {
	switch (error) {
		case UNRECOVERABLE_ERROR: 
			cerr << "Evoptool: unrecoverable error occured. Aborting." << endl;
			printLog((char*) __FILE__, __LINE__, (char*) "Unrecoverable error occured.", Evoptool::ERROR);
			printLog((char*) __FILE__, __LINE__, (char*) "Aborting.", Evoptool::ERROR);
			exit(-1);
			break;
	}
}
	
/* ------------------------------------------- Thread functions -------------------------------------------------- */

/* Initializer thread takes care about performing all the operations needed before run the new execution, it clean
 * folders, configures statistics and signals the execution beginning. */
void Evoptool::initializerThread() {
	printLog((char*) __FILE__, __LINE__, (char*) "New test initialization.", Evoptool::MESSAGE);
	if (!cleanTemporanyFiles()) signalError().emit(Evoptool::UNRECOVERABLE_ERROR);
	initializeTaskStructures();
	printLog((char*) __FILE__, __LINE__, (char*) "Parameters loaded.", Evoptool::MESSAGE);
	printLog((char*) __FILE__, __LINE__, (char*) "Configuring statistics computation.", Evoptool::MESSAGE);
	if (!configureStatistics()) signalError().emit(Evoptool::UNRECOVERABLE_ERROR);
	printLog((char*) __FILE__, __LINE__, (char*) "Statistics computation configured.", Evoptool::MESSAGE);	
	printLog((char*) __FILE__, __LINE__, (char*) "Running the test.", Evoptool::MESSAGE);

	signalRunTest().emit();
}

/* Runner thread takes care about creating, running and joining all the solver threads needed for the test 
 * execution. */
void Evoptool::runnerThread() {
	unsigned int i;
	Glib::Thread *th; 
	vector<Glib::Thread*> threads;
	ofstream file(STATUS_FILE, ios_base::trunc);
	
	file << "RUNNING";
	file.close();
	
	for (i = 0; i < _testParameters->solvers.size(); i++) {
		sigc::slot<void> s = sigc::bind(sigc::mem_fun(*this, &Evoptool::solverThread), _testParameters->solvers[i]);
		th = Glib::Thread::create(s, true);
		threads.push_back(th);
	}
	
	printLog((char*) __FILE__, __LINE__, (char*) "Resolutors started.", Evoptool::MESSAGE);
	
	for (i = 0; i < threads.size(); i++)
		threads[i]->join();
	threads.clear();
	
	printLog((char*) __FILE__, __LINE__, (char*) "Resolutors joined.", Evoptool::MESSAGE);
	signalFinalizeTest().emit();
}

/* Finalizer thread takes care about computing final statistics, exporting data if requested and cleaning up
 * TestParameters structure. */
void Evoptool::finalizerThread() {
	bool rval; 
	
	printLog((char*) __FILE__, __LINE__, (char*) "Test done.", Evoptool::MESSAGE);
	printLog((char*) __FILE__, __LINE__, (char*) "Finalizing test.", Evoptool::MESSAGE);	
	
	printLog((char*) __FILE__, __LINE__, (char*) "Computing final statistics.", Evoptool::MESSAGE);
	if (!computeStatistics()) {
		printLog((char*) __FILE__, __LINE__, (char*) "Error during statistics computing.", Evoptool::ERROR);
		signalError().emit(Evoptool::UNRECOVERABLE_ERROR);
	}
	printLog((char*) __FILE__, __LINE__, (char*) "Final statistics computed.", Evoptool::MESSAGE);
		
	if (_testParameters->exportData) {
		rval = exportDatas(_testParameters->destinationFolder, _testParameters->archiveName);
		if (!rval) signalError().emit(Evoptool::UNRECOVERABLE_ERROR);
	}
	
	printLog((char*) __FILE__, __LINE__, (char*) "Test finalized.", Evoptool::MESSAGE);
	printLog((char*) __FILE__, __LINE__, (char*) "Cleaning up tool's structures.", Evoptool::MESSAGE);
	
	deleteTestParameters();
	printLog((char*) __FILE__, __LINE__, (char*) "Execution done, exit.", Evoptool::MESSAGE);

	signalTestExecuted().emit();
}

/* This function defines the behavior of stopper thread which simply change the run condition varible. */
void Evoptool::stopperThread() {
	ofstream file(STATUS_FILE, ios_base::trunc);
	
	_testParameters->mutex->lock();
	_testParameters->runCondition = false; 
	file << "STOPPED";
	file.close();
	_testParameters->mutex->unlock();
	printLog((char*) __FILE__, __LINE__, (char*) "Forcing test execution end.", Evoptool::MESSAGE);
}

/* This function defines the behavior of the solver thread, that have simply to start the solver. */
void Evoptool::solverThread(Solver *solver) {
	solver->solve(_testParameters->mutex, &_testParameters->runCondition, _testParameters->maxGenerations);
}
	
/* ----------------------------------- Parameters file parser functions ------------------------------------------ */

/* This function initializes a part of TestParameters structure that can't be initialized by 
 * the XMLParser. */
void Evoptool::initTestParameters(TestParameters *params) {
	params->statisticsOptions = BASIC_GRAPHS;	
	params->rng = allocRNG();		
	params->mutex = new Glib::Mutex();
	params->runCondition = true;
}

/* ------------------------------- Tasks and Resolutors management functions -------------------------------------- */

/* Such some task problems need to save somewhere support files, this functions take care
 * about sinchronizes this operations with other Evoptool operations that can cause conflicts. 
 * (Ex: cleanTemporanyFile) */
void Evoptool::initializeTaskStructures() {
	ObjectiveFunction *obj = _testParameters->task;
	
	/* Texture restoration. */
	if (dynamic_cast<TextureRestoration*>(obj) != 0) {
		dynamic_cast<TextureRestoration*>(obj)->exportOptimalImage((char*) TEXTURE_RESTORATION_OPTIMAL_TMP);
		dynamic_cast<TextureRestoration*>(obj)->exportObservedImage((char*) TEXTURE_RESTORATION_OBSERVED_TMP);
	}
}

/* ----------------------------------- General purpose utility functions ------------------------------------------ */

/* Prints a new log line to log file and emit a NewLog signal in order to inform GUI if is present. 
 * Takes case about formatting style depending of the LogType.*/
void Evoptool::printLog(char *filename, int line, char *message, LogType type) {
	string log;
	ofstream logFile;
	const time_t timeVal = time(NULL);
	char *timeStr, tmp[10];
	
	timeStr = ctime(&timeVal);
	log.assign("[");
	log.append(timeStr, 11, 8);
	log.append("]");
	
	switch (type) {
		case MESSAGE: 
			log.append("[MESSAGE][");
			log.append(filename, 0, strcspn(filename, (char*) "."));
			log.append("]");
			break;
		case WARNING: 
			log.append("[WARNING][");		
			log.append(filename);
			log.append(" line: ");
			sprintf(tmp, "%d", line);
			log.append(tmp);
			log.append("]");
			break;
		case ERROR: 	
			log.append("[ERROR][");
			log.append(filename);
			log.append(" line: ");
			sprintf(tmp, "%d", line);
			log.append(tmp);
			log.append("]");		
			break;
	}
	
	log.append("[");	
	log.append(message);
	log.append("]\n");	
	
	logFile.open(LOG_FILE, ios_base::app);
	if (logFile.fail()) cerr << "Evoptool::printLog Error while opening log file." << endl;
	logFile << log;
	logFile.close();
	
	if (_mode == Evoptool::GUI) signalNewLog().emit(log);
	else cout << log;
}

/* Exports the entire content of temporany folder in a tar archive with a specified name and
 * to a specified folder. */
bool Evoptool::exportDatas(const char *destinationFolder, const char *archiveName) {
	stringstream cmd;
	int rval;
	
	cmd << "cd " << TMP_PATH << ";";
	cmd << "tar --exclude-vcs --exclude \"status*\" -cf ";
	cmd << destinationFolder << archiveName << ".tar *";
	
	rval = system(cmd.str().c_str());
	if (rval != 0) {
		printLog((char*) __FILE__, __LINE__, (char*) "Error while exporting data", Evoptool::WARNING);
		return false;
	}
	
	printLog((char*) __FILE__, __LINE__, (char*) "Data exported.", Evoptool::MESSAGE);
	return true;

}

/* Cleans up the temporany folder from data and images. Deletes everything except content of ../temp/support/ 
 * folder. */
bool Evoptool::cleanTemporanyFiles() {
	int rval;
	stringstream cmd1, cmd2;
	
	/* Delete temporany data. */
	cmd1 << "rm -rf " << TMP_DATA_PATH << "*";
	rval = system(cmd1.str().c_str());
	if (rval != 0) {
		printLog((char*) __FILE__, __LINE__, (char*) "Error while deleting temporany data.", Evoptool::WARNING);
		return false;
	}
	
	/* Delete temporany images. */
	cmd2 << "rm -rf " << TMP_IMAGE_PATH << "*";
	rval = system(cmd2.str().c_str());
	if (rval != 0) {
		printLog((char*) __FILE__, __LINE__, (char*) "Error while deleting temporany images.", Evoptool::WARNING);
		return false;
	}
	
	printLog((char*) __FILE__, __LINE__, (char*) "Temporany folder cleaned.", Evoptool::MESSAGE);
	return true;
}

/* Takes care about freeing memory allocated for TestParameters structure. */
void Evoptool::deleteTestParameters() {
	unsigned int i; 
	
	freeRNG(_testParameters->rng);
	delete _testParameters->task;
	for (i = 0; i < _testParameters->solvers.size(); i++) delete _testParameters->solvers[i];
	_testParameters->solvers.clear();
	delete _testParameters->mutex;
	delete _testParameters;
}

/* ---------------------------------------- Statistics managment functions -------------------------------------------- */

/* It performs all the operations needed to configuring statistics computation and visualization. */
bool Evoptool::configureStatistics() { 
	
	/* Configure basic graph plotting as average, best and worst instance, fitness calls. */
	if (_testParameters->statisticsOptions & BASIC_GRAPHS) {
		if (!configureBasicGraphsStatistics()) {
			printLog((char*) __FILE__, __LINE__, (char*) "Basic graphs not configured.", Evoptool::WARNING);
			return false;
		}
		printLog((char*) __FILE__, __LINE__, (char*) "Basic graphs configured.", Evoptool::MESSAGE);
	}
	
	if (_testParameters->statisticsOptions & BEST_AS_IMAGE) {
		if (!configureBestAsImageStatistics()) {
			printLog((char*) __FILE__, __LINE__, (char*) "Best as image statistics not configured.", Evoptool::WARNING);
			return false;
		}
		printLog((char*) __FILE__, __LINE__, (char*) "Best as image statistics configured.", Evoptool::MESSAGE);
	}
	
	return true;
}

/* Takes care about configuring basic graphs statistics. */
bool Evoptool::configureBasicGraphsStatistics() {
	unsigned int i, j;
	unsigned const int algorithmsNum = _testParameters->solvers.size();
	string algorithm[algorithmsNum];
	string graphType[4] = {"1:2", "1:3", "1:4", "5:2"};
	string data[4] = {"plot ", "plot", "plot", "plot"};
	string file[4] = {AVG_GNUPLOT_FILE, BEST_GNUPLOT_FILE , WORST_GNUPLOT_FILE, FCALLS_GNUPLOT_FILE};
	
	/* Setup structures. */
	for (i = 0; i < algorithmsNum; i++)
		algorithm[i] = _testParameters->solvers[i]->getAlgorithmName();
	
	/* Building specific data strings. */
	for (i = 0; i < 4; i++) {
		for (j = 0; j < algorithmsNum; j++) {
			stringstream sstream;
			
			sstream << "\"" << TMP_DATA_PATH << algorithm[j] << ".dat\" using ";
			sstream << graphType[i] << " with lines lw 2 lt " << j + 1 << " title \"";
			sstream << algorithm[j] << "\", ";
			data[i].append(sstream.str());
		}
		data[i].erase(data[i].size()  - 2, 2);
	}
	
	/* Configure gnuplot files. */
	for (i = 0; i < 4; i++) {
		if (!setGnuplotData(file[i].c_str(), data[i])) return false;
		if (!setGnuplotSize(file[i].c_str(), GRAPH_HEIGHT, GRAPH_WIDTH)) return false;
		if (i < 3) {
			if (!setGnuplotXRange(file[i].c_str(), _testParameters->maxGenerations)) return false;
		}
	}
	
	return true;
}

/* Takes care about configuring best as image statistics support variables. */
bool Evoptool::configureBestAsImageStatistics() {
	unsigned int i, fnum = _testParameters->solvers.size();

	if (_bestAsImageFilesAllocated) {
		for (i = 0; i < _bestAsImageFilesNum; i++)
			delete [] _bestAsImageFiles[i];
		delete [] _bestAsImageFiles;
	}
	
	_bestAsImageFiles = new char*[fnum];
	_bestAsImageFilesNum = fnum;
	_bestAsImageFilesAllocated = true;
	_bestAsImageBlockSize = _testParameters->task->getSize();
	 
	for (i = 0; i < fnum; i++) {
		string name = _testParameters->solvers[i]->getAlgorithmName();
		_bestAsImageFiles[i] = new char[100];
		sprintf(_bestAsImageFiles[i], "%sBI_%s.dat", TMP_DATA_PATH, name.c_str());
	}
	
	return true;
}

/* It computes the statistics results. If no argument is passed, it computes all the statistics setted
 * for the current test execution.*/
bool Evoptool::computeStatistics(unsigned int options, unsigned int height, unsigned int width) {
	
	if (options == 0) options = _testParameters->statisticsOptions;
	
	if (options & AVG_GRAPH) {
		if (!setGnuplotSize(AVG_GNUPLOT_FILE, height, width)) return false;
		if (!callGnuplot(AVG_GNUPLOT_FILE)) return false;
	}
	
	if (options & BEST_GRAPH) { 
		if (!setGnuplotSize(BEST_GNUPLOT_FILE, height, width)) return false;
		if (!callGnuplot(BEST_GNUPLOT_FILE)) return false;
	}

	if (options & WORST_GRAPH) { 
		if (!setGnuplotSize(WORST_GNUPLOT_FILE, height, width)) return false;
		if (!callGnuplot(WORST_GNUPLOT_FILE)) return false;
	}
	
	if (options & FCALLS_GRAPH) { 
		if (!setGnuplotSize(FCALLS_GNUPLOT_FILE, height, width)) return false;
		if (!callGnuplot(FCALLS_GNUPLOT_FILE)) return false;
	}
	
	if (options & BEST_AS_IMAGE) { 
		if (!convertBestToImage(const_cast<const char **>(_bestAsImageFiles), _bestAsImageFilesNum,TMP_IMAGE_PATH)) 
			return false;
	}
	return true;
}

/* This function converts best individuals to an image representation. Best individuals are stored in binary 
 * files specified by the parameters, and created images will be saved to the destination folder parameter. */
bool Evoptool::convertBestToImage(const char **files, int fileNumber, const char *destination) {
	int i, fileLen;
	char *best, tmp[100];
	ifstream file;
	struct stat fileStat;
	
	for (i = 0; i < fileNumber; i++) {
		/* Obtains file length. */
		if (stat(files[i], &fileStat) == 0)
        		fileLen = fileStat.st_size;
    		else {
    			printLog((char*) __FILE__, __LINE__, (char*) "Error while allocation of stat structure.", Evoptool::WARNING);
    			return false;
    		}
    		
    		/* Checks if there are enought data. */
    		if (fileLen < (int) _bestAsImageBlockSize * 2) {
    			printLog((char*) __FILE__, __LINE__, (char*) "Not enought data to read.", Evoptool::WARNING);
    			return false;
    		}

		/* Opens the file */
		file.open(files[i], ios_base::in | ios_base::binary);
		if (file.fail()) {
			printLog((char*) __FILE__, __LINE__, (char*) "Failed to open best as image file.", Evoptool::WARNING);
			return false;
		}
		
		/* Computes read position. */
		if ( fileLen % _bestAsImageBlockSize == 0) {
			int tmp = fileLen / _bestAsImageBlockSize;
			file.seekg( (tmp - 2) * _bestAsImageBlockSize);
		} else {
			int tmp = (fileLen - (fileLen % _bestAsImageBlockSize)) / _bestAsImageBlockSize;
			file.seekg( (tmp - 3) * _bestAsImageBlockSize);
		}
		
		/* Converts to image. */
		best = new char[_bestAsImageBlockSize];
		file.read(best, _bestAsImageBlockSize);
		if (file.gcount() != (int) _bestAsImageBlockSize) {
			printLog((char*) __FILE__, __LINE__, (char*) "Too less bytes readed.", Evoptool::WARNING);
			return false;
		}
		
		sprintf(tmp, "%s%d.png", TEXTURE_RESTORATION_TMP, i);
		GraphicsTool::binaryInstanceToImage(best, _bestAsImageBlockSize, tmp);
		
		/* Closes and releases resources. */
		file.close(); 
		delete[] best;
	}
	
	return true;
}
/* ---------------------------------------- Gnuplot utility functions -------------------------------------------- */

/* Reads the specified gnuplot configuration file till it reach the data section, it procedes by 
 * deleting previous data and writing new one. */
bool Evoptool::setGnuplotData(const char *file, string data) {
	FILE* cfile;
	char tmp[256];
	int count, pos;
	
	cfile = fopen(file, "r+");
	if (cfile == NULL) {
		printLog((char*) __FILE__, __LINE__, (char*) "Failed to open gnuplot configuration file.", Evoptool::WARNING);
		return false;
	}
	
	do {
		count = fscanf(cfile, "%s", tmp);
		if (count == EOF) {
			printLog((char*) __FILE__, __LINE__, (char*) "EOF reached without find data section.", Evoptool::WARNING);
			return false;
		}
	} while (strcmp(tmp, "#PLOT-DATA-SECTION#") != 0);
	
	count = fprintf(cfile, "%c", '\n');
	count = fprintf(cfile, "%s", data.c_str());
	
	if (count == -1) {
		printLog((char*) __FILE__, __LINE__, (char*) "Failed to write data section.", Evoptool::WARNING);
		return false;
	}
	
	pos = ftell(cfile);
	fclose(cfile);
	
	count = truncate(file, pos);
	if (count != 0) {
		printLog((char*) __FILE__, __LINE__, (char*) "Failed to truncate gnuplot file.", Evoptool::WARNING);
		return false;
	}
	
	printLog((char*) __FILE__, __LINE__, (char*) "Gnuplot data configured", Evoptool::MESSAGE);
	return true;
}

/* Reads the specified gnuplot configuration file till it reach the image size section, than it
 * set the new image size according to the arguments and with a minimum size of (100,100). */
bool Evoptool::setGnuplotSize(const char *file, int height, int width) {
	FILE* cfile;
	char tmp[256];
	int control;
	
	if (height < 100) height = 100;
	if (width < 100) width = 100;
	
	cfile = fopen(file, "r+");
	if (cfile == NULL) {
		printLog((char*) __FILE__, __LINE__, (char*) "Failed to open gnuplot configuration file.", Evoptool::WARNING);
		return false;
	}
	
	do {
		control = fscanf(cfile, "%s", tmp);
		if (control == EOF) {
			printLog((char*) __FILE__, __LINE__, (char*) "EOF reached without find image size section.", Evoptool::WARNING);
			return false;
		}
	} while (strcmp(tmp, "#IMAGE-SIZE#") != 0);
	
	fprintf(cfile, "%c", '\n');
	control = fprintf(cfile, "set term png enhanced size %d,%d\n", width, height);
	
	if (control == -1) {
		printLog((char*) __FILE__, __LINE__, (char*) "Failed to write image size section.", Evoptool::WARNING);
		return false;
	}
	
	fclose(cfile);
	
	return true;
}

/* Reads the specified gnuplot configuration file till it reach the xrange section, than it
 * set the new xrange according to the argument. */
bool Evoptool::setGnuplotXRange(const char *file, int range) {
	FILE* cfile;
	char tmp[256];
	int control;
	
	cfile = fopen(file, "r+");
	if (cfile == NULL) {
		printLog((char*) __FILE__, __LINE__, (char*) "Failed to open gnuplot configuration file.", Evoptool::WARNING);
		return false;
	}
	
	do {
		control = fscanf(cfile, "%s", tmp);
		if (control == EOF) {
			printLog((char*) __FILE__, __LINE__, (char*) "EOF reached without find xrange section.", Evoptool::WARNING);
			return false;
		}
	} while (strcmp(tmp, "#XRANGE#") != 0);
	
	fprintf(cfile, "%c", '\n');
	if (range == -1) control = fprintf(cfile, "set xrange [] noreverse writeback\n");
	else control = fprintf(cfile, "set xrange [0:%d] noreverse nowriteback\n", range);
	
	if (control == -1) {
		printLog((char*) __FILE__, __LINE__, (char*) "Failed to write xrange section." , Evoptool::WARNING);
		return false;
	}
	
	fclose(cfile);
	
	return true;
}

/* This functions is a wrapper for gnuplot calls. */
bool Evoptool::callGnuplot(const char *file) {
	stringstream cmd;
	int rval;
	
	cmd << "gnuplot " << file;
	rval = system(cmd.str().c_str());
	if (rval != 0) {
		printLog((char*) __FILE__, __LINE__, (char*) "Call to gnuplot failed.", Evoptool::WARNING);
		return false;
	}
	
	return true;
}
	
/* ------------------------------- Random number generator utility functions ------------------------------------- */

/* Allocates and initialiazes a random number generator with a pseudo-random seed. */
gsl_rng* Evoptool::Evoptool::allocRNG() {
	const gsl_rng_type *T;
	gsl_rng *r;
	
	gsl_rng_env_setup();
	T = gsl_rng_default;
	r = gsl_rng_alloc(T);
	gsl_rng_set(r,random_seed());
	
	printLog((char*) __FILE__, __LINE__, (char*) "Random number generator initialized.", Evoptool::MESSAGE);
	return r;
}

/* Deallocates memory associated with given random number generator. */
void Evoptool::freeRNG(gsl_rng *r) {
	printLog((char*) __FILE__, __LINE__, (char*) "Freeing random number generator.", Evoptool::MESSAGE);
	gsl_rng_free(r);
}
