
/*********************************************************************************
 *  Copyright© 2010 Gabriele Valentini (gabriele.valentini@mail.polimi.it)	 *
 *										 *
 *  This file is part of Evoptool.						 *
 * 										 *
 *  Evptool is free software: you can redistribute it and/or modify              *
 *  it under the terms of the GNU General Public License as published by         *
 *  the Free Software Foundation, either version 3 of the License, or            *
 *  (at your option) any later version.                                          *
 *										 *
 *  Evoptool is distributed in the hope that it will be useful,                  *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of  		 *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 		 *
 *  GNU General Public License for more details.				 *
 *										 *
 *  You should have received a copy of the GNU General Public License		 *
 *  along with Evoptool.  If not, see <http://www.gnu.org/licenses/>.    	 *	
 *********************************************************************************/

#include "TextureRestoration.h"
#include "GraphicsTool.h"
#include "path.h"
#include <math.h>
#include <fstream>

/* Class constructor. */
TextureRestoration::TextureRestoration(const char *filename) : ObjectiveFunction(1), _fileName(filename) {
	init(filename);
}

/* Class destructor. */
TextureRestoration::~TextureRestoration() {
	delete _observed;
	delete _optimal;
	delete[] _structure;
	gsl_rng_free(_r);
}

/* Fitness function. */
double TextureRestoration::f(BinaryInstance* bi) {
	volatile double h = 0, tmp1 = 0, tmp2 = 0;
	int i;
	
	switch(_noise) {
		case CHANNEL_NOISE: 
			for (i = 0; i < _structureLength; i++) {
				tmp1 += bi->getHarmonicEncoding(_structure[i][0]) * bi->getHarmonicEncoding(_structure[i][1]);
				if (i < _size) {
					tmp2 += bi->getHarmonicEncoding(i) * _observed->getHarmonicEncoding(i);
				} 
			}	
			h = -1.0 * BETA * tmp1 - 0.5 * tmp2 * log((1 - PNOISE) / PNOISE);
			break; 
			
		case WHITE_NOISE: break;
	}
	
	if (_normalize) {
		h = h * 100.0 / _optimalValue;
		if (h > 100.0) h = 100.0;
		if (h < 0.0) h = 0.0;
	}

	return h;
}

/* Returns the filename of this objective function instances. */
string TextureRestoration::fileName() {
	return _fileName;
}

/* Exports optimal individual as image to the specified file. */
void TextureRestoration::exportOptimalImage(char *file) {
	GraphicsTool::binaryInstanceToImage(_optimal, file);
}

/* Exports observed individual as image to the specified file. */
void TextureRestoration::exportObservedImage(char *file) {
	GraphicsTool::binaryInstanceToImage(_observed, file);
}
	
/* Initialize the structures needed by the fitness function. */
void TextureRestoration::init(const char *filename) {
	ofstream textureRestorationFile;
	const gsl_rng_type *T;
	char fn[100];
	
	strcpy(fn, filename);
	
	gsl_rng_env_setup();
	T = gsl_rng_default;
	_r = gsl_rng_alloc(T);
	gsl_rng_set(_r, time(NULL));
	
	_optimal = GraphicsTool::imageToBinaryInstance(fn);
	_size = _optimal->getSize();
	loadStructure();
	_noise = CHANNEL_NOISE;
	_observed = addNoise(_optimal);
	_normalize = false;
	_optimalValue = f(_optimal);
	_normalize = true;
	
	strcpy(fn, TEXTURE_RESTORATION_BIN);
	textureRestorationFile.open(fn, ios_base::out | ios_base::binary);
	if (!textureRestorationFile.is_open()) cerr << "[TextureRestoration] cannot open destination file!" << endl;
	textureRestorationFile.write(_optimal->toCharArray(), _optimal->getSize());
	textureRestorationFile.write(_observed->toCharArray(), _observed->getSize());
	textureRestorationFile.close();
}

/* Corrupts the image binary representation by adding somekind of noise. */
BinaryInstance* TextureRestoration::addNoise(BinaryInstance* original) {
	int index;
	BinaryInstance *observed = new BinaryInstance(*original);
	
	switch(_noise) {
		/* Changes indipendentelly every variable value with a fixed probability. */
		case CHANNEL_NOISE: 
			for (int i = 0; i < observed->getSize(); i++) {
				if (gsl_ran_bernoulli(_r, PNOISE)) {
					index = observed->get(i);
					if (index == 1) observed->set(i,0);
					else observed->set(i,1);
				}
			}
			break; 
			
		case WHITE_NOISE: break;
	}
	return observed;
}

/* Load the structure used to compute the fitness value. */
void TextureRestoration::loadStructure() {
	int i, j, rsize, ssize, count = 0;
	ssize = sqrt(_size);
	rsize = ssize - 1;
	_structureLength = 3 * pow(rsize, 2) + 2 * (rsize);
	_structure = new int *[_structureLength];
	
	/* Initialize inner matrix. */
	for (i = 0; i < rsize; i++) {
		for (j = 0; j < rsize; j++) {
			_structure[count] = new int [2];
			_structure[count][0] = i * ssize + j;
			_structure[count][1] = i * ssize + j + 1;
			count++;
			
			_structure[count] = new int [2];
			_structure[count][0] = i * ssize + j;
			_structure[count][1] = i * ssize + j + ssize;
			count++;
			
			_structure[count] = new int [2];
			_structure[count][0] = i * ssize + j;
			_structure[count][1] = i * ssize + j + ssize + 1;
			count++; 
		}
	}
	
	/* Initialize matrix left side. */
	j = rsize;
	for (i = 0; i < rsize; i++) {
		_structure[count] = new int [2];
		_structure[count][0] = i * ssize + j;
		_structure[count][1] = i * ssize + j + ssize;
		count++;
	}
	
	/* Initialize matrix down side. */ 
	i = rsize;
	for (j = 0; j < rsize; j++) {
		_structure[count] = new int [2];
		_structure[count][0] = i * ssize + j;
		_structure[count][1] = i * ssize + j + 1;
		count++;
	}
}
